USE [common_db]
GO

/*
	Definitions of all tables in database
*/
EXECUTE [dbo].[DropObjectIfExists] 'tables', 'U'
GO
CREATE TABLE [dbo].[tables] (
	[id]			BIGINT IDENTITY PRIMARY KEY,
	[id_global]		BIGINT NOT NULL DEFAULT 1,
	[title]			NVARCHAR(255) NOT NULL,
	[description]	NTEXT,
	CONSTRAINT [fk_tables_global_ref] 
		FOREIGN KEY ([id_global])
		REFERENCES [dbo].[objects] ([id])
)
GO

/*
	Returns global identifier of main table ('table' table')	
*/
EXECUTE [dbo].[DropObjectIfExists] 'f_GetRootTableId', 'FN'
GO
CREATE FUNCTION [dbo].[f_GetRootTableId] (
	)
RETURNS BIGINT
AS
BEGIN
	DECLARE @root_t BIGINT
	SELECT @root_t = [id] FROM [dbo].[objects] WHERE [id_table] = 1 AND [id_row] = 1
	RETURN @root_t
END
GO

/*
	Returns global identifier of object by table identifier and row identifier
*/
EXECUTE [dbo].[DropObjectIfExists] 'f_GetGlobalId', 'FN'
GO
CREATE FUNCTION [dbo].[f_GetGlobalId] (
	@id_table BIGINT, 
	@id_row BIGINT
	)
RETURNS BIGINT
AS
BEGIN
	DECLARE @global_id BIGINT
	SET @global_id = -1
	SELECT @global_id = [id] FROM [dbo].[objects] WHERE [id_table] = @id_table AND [id_row] = @id_row
	RETURN @global_id
END
GO

/*
	first DELETE trigger for [tables]
	Delete references to table in [objects_relations] and [objects]
*/
EXECUTE [dbo].[DropObjectIfExists] 'trig_delete_table_global_ref', 'TR'
GO
CREATE TRIGGER [trig_delete_table_global_ref] ON [dbo].[tables]
FOR DELETE
AS
BEGIN
	DECLARE @deleted_id BIGINT
	SELECT @deleted_id = [id_global] FROM DELETED
	DELETE FROM [dbo].[tables_fields] WHERE [id_table] = @deleted_id
	DELETE FROM [dbo].[objects_relations] WHERE [id_object] = @deleted_id
	DELETE FROM [dbo].[objects_relations] WHERE [id_child] = @deleted_id
	DELETE FROM [dbo].[objects] WHERE [id] = @deleted_id	
END
EXECUTE [dbo].[sp_settriggerorder] 'trig_delete_table_global_ref', 'first', 'delete'
GO

/*
	last trigger for [tables]
	Drops deleted table from database. This trigger works after [trig_delete_table_global_ref]
*/
EXECUTE [dbo].[DropObjectIfExists] 'trig_delete_table', 'TR'
GO
CREATE TRIGGER [trig_delete_table] ON [dbo].[tables]
FOR DELETE
AS
BEGIN
	DECLARE @del_title NVARCHAR(255)
	SELECT @del_title = [title]  FROM DELETED
	DECLARE @drop_st NVARCHAR(255)
	SET @drop_st N'DELETE FROM [dbo].[' + @del_title + N']'
	EXECUTE @drop_st
	SET @drop_st = N'DROP TABLE [dbo].[' + @del_title + N']'
	EXECUTE [dbo].[sp_executesql] @drop_st
END
EXECUTE [dbo].[sp_settriggerorder] 'trig_delete_table', 'last', 'delete'
GO

/*
	First INSERT trigger for [tables]
	Inserts global reference to [objects], updates [id_global] and make inserted table as child for root table
*/
EXECUTE [dbo].[DropObjectIfExists] 'trig_insert_table_add_global_ref', 'TR'
GO
CREATE TRIGGER [trig_insert_table_add_global_ref] ON [dbo].[tables]
FOR INSERT
AS
BEGIN
	DECLARE @ins_id BIGINT
	DECLARE @ins_descr NVARCHAR(255)
	SELECT @ins_id = [id], @ins_descr = [title] FROM INSERTED	
	SET @ins_descr = '[' + @ins_descr + '] table'
	INSERT INTO [dbo].[objects] ([id_table],[id_row],[description]) VALUES (1, @ins_id, @ins_descr)	
	DECLARE @global_id BIGINT
	SELECT @global_id = dbo.f_GetGlobalId(1, @ins_id)	
	UPDATE [dbo].[tables] SET [id_global] = @global_id WHERE [id] = @ins_id	
	INSERT INTO [dbo].[objects_relations] ([id_object],[id_child]) VALUES (dbo.f_GetRootTableId(), @global_id)
END
GO
EXECUTE [dbo].[sp_settriggerorder] 'trig_insert_table_add_global_ref', 'first', 'insert'
GO

/*
	[tables] table. Trigger 'trig_insert_table_create_table' must be created AFTER this query!
*/
INSERT INTO [dbo].[tables] ([title], [description]) VALUES ('tables', 'contain definitions for all tables in database')
GO

/*
	Last INSERT trigger for [tables]. This trigger must work after 'trig_insert_table_add_global_ref'
	In this trigger:
	1. Create new table named [title] in database.
	2. Create [f_GetRoot_TABLENAME] and [f_GetTableId_TABLENAME] for getting global and local identifiers of created table
	3. Create delete and insert triggers for created table
*/
EXECUTE [dbo].[DropObjectIfExists] 'trig_insert_table_create_table', 'TR'
GO
CREATE TRIGGER [trig_insert_table_create_table] ON [dbo].[tables]
FOR INSERT
AS
BEGIN
	DECLARE @ins_id BIGINT
	DECLARE @ins_title NVARCHAR(255)
	SELECT @ins_id = [id], @ins_title = [title] FROM INSERTED
	DECLARE @st_create NVARCHAR(500)
	SET @st_create = N'EXECUTE [dbo].[DropObjectIfExists] ''' + @ins_title + ''', ''U'''	
	EXECUTE [dbo].[sp_executesql] @st_create
	SET	@st_create =
		N'CREATE TABLE [dbo].[' + @ins_title + N'](
			[id] BIGINT IDENTITY PRIMARY KEY, 
			[id_global] BIGINT NOT NULL DEFAULT 1,			
			[title] NVARCHAR(255),
			CONSTRAINT [fk_' + @ins_title + N'_global_ref] 
				FOREIGN KEY ([id_global])
				REFERENCES [dbo].[objects] ([id])
			 )'		
	EXECUTE [dbo].[sp_executesql] @st_create
	DECLARE @getroot_f NVARCHAR(255)
	SET @getroot_f = N'f_GetRoot_' + @ins_title
	DECLARE @st_getroot_f NVARCHAR(1000)	
	SET @st_getroot_f = N'EXECUTE [dbo].[DropObjectIfExists] ''' + @getroot_f + ''', ''FN'''	
	EXECUTE [dbo].[sp_executesql] @st_getroot_f
	SET @st_getroot_f = 
		N'CREATE FUNCTION ' + @getroot_f + N'() 
		RETURNS BIGINT 
		AS 
		BEGIN 
			DECLARE @id_root BIGINT 
			SELECT @id_root = [id_global] FROM [dbo].[tables] WHERE [title] = ''' + @ins_title + ''' ' +
			N'RETURN @id_root
		END'
	EXECUTE [dbo].[sp_executesql] @st_getroot_f
	DECLARE @gettabid_f NVARCHAR(255)
	SET @gettabid_f = N'f_GetTableId_' + @ins_title
	DECLARE @st_gettabid_f NVARCHAR(1000)
	SET @st_gettabid_f = N'EXECUTE [dbo].[DropObjectIfExists] ''' + @gettabid_f + ''', ''FN'''	
	EXECUTE [dbo].[sp_executesql] @st_gettabid_f
	SET @st_getroot_f = 
		N'CREATE FUNCTION ' + @gettabid_f + N'() 
		RETURNS BIGINT 
		AS 
		BEGIN 
			DECLARE @id_table BIGINT 
			SELECT @id_table = [id] FROM [dbo].[tables] WHERE [title] = ''' + @ins_title + ''' ' +
			N'RETURN @id_table
		END'
	EXECUTE [dbo].[sp_executesql] @st_getroot_f
	DECLARE @trig_insert NVARCHAR(255)
	SET @trig_insert = N'[trig_insert_' + @ins_title + N'_add_global_ref]'
	DECLARE @st_trig_insert NVARCHAR(2000)		
	SET @st_trig_insert = 
	N'CREATE TRIGGER ' + @trig_insert + ' ON [dbo].[' + @ins_title + ']  
	FOR INSERT 
	AS
	BEGIN
		DECLARE @ins_id BIGINT
		DECLARE @ins_title NVARCHAR(255)
		DECLARE @table_name NVARCHAR(255)
		DECLARE @descr NVARCHAR(255)
		SET @table_name = ''' + @ins_title + N'''		
		SELECT @ins_id = [id], @ins_title = [title] FROM INSERTED						
		SET @descr = @ins_title + '' in '' + @table_name
		DECLARE @root_id BIGINT
		SELECT @root_id = [dbo].[' + @getroot_f + N']()		
		DECLARE @table_id BIGINT
		SELECT @table_id = [dbo].[' + @gettabid_f + N']()					
		INSERT INTO [dbo].[objects] ([id_table],[id_row],[description]) VALUES (@table_id, @ins_id,@descr)
		DECLARE @global_id BIGINT	
		SELECT @global_id = [dbo].[f_GetGlobalId](@table_id,@ins_id)		
		UPDATE [' + @ins_title + N'] SET [id_global] = @global_id WHERE [id] = @ins_id
		INSERT INTO [dbo].[objects_relations] ([id_object],[id_child]) VALUES (@root_id,@global_id)
	END'	
	EXECUTE [dbo].[sp_executesql] @st_trig_insert
	DECLARE @trig_delete NVARCHAR(255)
	SET @trig_delete = N'[trig_delete_' + @ins_title + N'_delete_global_ref]'
	DECLARE @st_trig_delete NVARCHAR(2000)
	SET @st_trig_delete = 
	N'CREATE TRIGGER ' + @trig_delete + ' ON [dbo].[' + @ins_title + ']
	FOR DELETE
	AS
	BEGIN
		DECLARE @deleted_id BIGINT	
		SELECT @deleted_id = [id_global] FROM DELETED
		DELETE FROM [dbo].[objects_relations] WHERE [id_object] = @deleted_id
		DELETE FROM [dbo].[objects_relations] WHERE [id_child] = @deleted_id
		DELETE FROM [dbo].[objects] WHERE [id] = @deleted_id
	END'
	EXECUTE [dbo].[sp_executesql] @st_trig_delete
END
GO
EXECUTE [dbo].[sp_settriggerorder] 'trig_insert_table_create_table', 'last', 'insert'
GO

/*
	Returns table identifier (local, global) by table name
*/
EXECUTE [dbo].[DropObjectIfExists] 'f_GetTableId', 'FN'
GO
CREATE FUNCTION [dbo].[f_GetTableId] (
	@table_name NVARCHAR(255)
	)
RETURNS BIGINT
AS
BEGIN
	DECLARE @ret BIGINT
	SET @ret = -1
	SELECT @ret = [id] FROM [dbo].[tables] WHERE [title] = @table_name
	RETURN @ret
END
GO

EXECUTE [dbo].[DropObjectIfExists] 'f_GetGlobalTableId', 'FN'
GO
CREATE FUNCTION [dbo].[f_GetGlobalTableId] (
	@table_name NVARCHAR(255)
	)
RETURNS BIGINT
AS
BEGIN
	DECLARE @ret BIGINT
	SET @ret = -1
	SELECT @ret = [id_global] FROM [dbo].[tables] WHERE [title] = @table_name
	RETURN @ret
END
GO

EXECUTE [dbo].[DropObjectIfExists] 'f_GetTableName', 'FN'
GO
CREATE FUNCTION [dbo].[f_GetTableName] (
	@table_id BIGINT
	)
RETURNS NVARCHAR(255)
AS
BEGIN
	DECLARE @ret NVARCHAR(255)
	SET @ret = ''
	SELECT @ret = [title] FROM [dbo].[tables] WHERE [id] = @table_id
	RETURN @ret
END
GO
/*
	TABLES_FIELDS
*/
INSERT INTO [dbo].[tables] ([title],[description]) VALUES ('tables_fields','definitions of tables fields')
GO
ALTER TABLE [dbo].[tables_fields] ADD [id_table] BIGINT NOT NULL DEFAULT 1
GO
ALTER TABLE [dbo].[tables_fields] ADD CONSTRAINT [fk_field_table_id]
	FOREIGN KEY ([id_table])
	REFERENCES [dbo].[tables] ([id])
GO

INSERT INTO [dbo].[tables] ([title],[description]) VALUES ('data_types_categories', 'definitions of data types categories')
GO
ALTER TABLE [dbo].[data_types_categories] ADD [description] NTEXT DEFAULT NULL
GO

INSERT INTO [dbo].[data_types_categories] ([title],[description]) VALUES (N'exact_numbers', N'Exact numbers')
INSERT INTO [dbo].[data_types_categories] ([title],[description]) VALUES (N'unicode_strings', N'Unicode character strings')
INSERT INTO [dbo].[data_types_categories] ([title],[description]) VALUES (N'strings', N'Simples character strings')
INSERT INTO [dbo].[data_types_categories] ([title],[description]) VALUES (N'approximate_numbers', N'Approximate numbers')
INSERT INTO [dbo].[data_types_categories] ([title],[description]) VALUES (N'binary_data', N'Binary data')
INSERT INTO [dbo].[data_types_categories] ([title],[description]) VALUES (N'date_time', N'Date and time')
INSERT INTO [dbo].[data_types_categories] ([title],[description]) VALUES (N'other', N'Other types of data')
INSERT INTO [dbo].[data_types_categories] ([title],[description]) VALUES (N'references', N'Object references')
GO

EXECUTE [dbo].[DropObjectIfExists] 'f_GetDataTypeCategoryId', 'FN'
GO
CREATE FUNCTION [dbo].[f_GetDataTypeCategoryId] (
	@data_type NVARCHAR(255)
	)
RETURNS BIGINT
AS
BEGIN
	DECLARE @ret BIGINT
	SET @ret = -1
	SELECT @ret = [id] FROM [dbo].[data_types_categories] WHERE [title] = @data_type
	RETURN @ret
END
GO

EXECUTE [dbo].[DropObjectIfExists] 'f_GetDataTypeCategoryName', 'FN'
GO
CREATE FUNCTION [dbo].[f_GetDataTypeCategoryName] (
	@data_type BIGINT
	)
RETURNS NVARCHAR(255)
AS
BEGIN
	DECLARE @ret NVARCHAR(255)
	SET @ret = ''
	SELECT @ret = [title] FROM [dbo].[data_types_categories] WHERE [id] = @data_type
	RETURN @ret
END
GO

INSERT INTO [dbo].[tables] ([title],[description]) VALUES ('data_types', 'Definitions of data types')
GO
ALTER TABLE [dbo].[data_types] ADD [id_data_type_category] BIGINT NOT NULL
GO
ALTER TABLE [dbo].[data_types] ADD CONSTRAINT [fk_data_type_category] 
	FOREIGN KEY ([id_data_type_category])
	REFERENCES [dbo].[data_types_categories] ([id])
GO
ALTER TABLE [dbo].[data_types] ADD [description] NTEXT DEFAULT NULL
GO

EXECUTE [dbo].[DropObjectIfExists] 'InsertDataType', 'P'
GO
CREATE PROCEDURE [dbo].[InsertDataType] 
	@data_type_cat NVARCHAR(255), 
	@data_type NVARCHAR(255),
	@descr NVARCHAR(1000)
AS
BEGIN
	DECLARE @cat_id BIGINT
	SELECT @cat_id = [dbo].[f_GetDataTypeCategoryId](@data_type_cat)
	INSERT INTO [dbo].[data_types] ([title],[description],[id_data_type_category]) 
			VALUES(@data_type, @descr, @cat_id)
	RETURN 0
END
GO

EXECUTE [dbo].[InsertDataType] N'exact_numbers', N'BIGINT', N'Big integer type'
EXECUTE [dbo].[InsertDataType] N'exact_numbers', N'NUMERIC', N'Numeric type'
EXECUTE [dbo].[InsertDataType] N'exact_numbers', N'BIT', N'Bit (boolean) type'
EXECUTE [dbo].[InsertDataType] N'exact_numbers', N'SMALLINT', N'Small integer type'
EXECUTE [dbo].[InsertDataType] N'exact_numbers', N'DECIMAL', N'Decimal type'
EXECUTE [dbo].[InsertDataType] N'exact_numbers', N'SMALLMONEY', N'Small money type'
EXECUTE [dbo].[InsertDataType] N'exact_numbers', N'INT', N'Simple integer type'
EXECUTE [dbo].[InsertDataType] N'exact_numbers', N'TINYINT', N'Tiny integer type'
EXECUTE [dbo].[InsertDataType] N'exact_numbers', N'MONEY', N'Simple money type'
GO

EXECUTE [dbo].[InsertDataType] N'unicode_strings', N'NCHAR', N'Unicode character type'
EXECUTE [dbo].[InsertDataType] N'unicode_strings', N'NTEXT', N'Unicode text type'
EXECUTE [dbo].[InsertDataType] N'unicode_strings', N'NVARCHAR', N'Unicode variable character type'
GO

EXECUTE [dbo].[InsertDataType] N'strings', N'CHAR', N'Simple character type'
EXECUTE [dbo].[InsertDataType] N'strings', N'TEXT', N'Simple text type'
EXECUTE [dbo].[InsertDataType] N'strings', N'VARCHAR', N'Simple variable character type'
GO

EXECUTE [dbo].[InsertDataType] N'approximate_numbers', N'FLOAT', N'Floating point number type'
EXECUTE [dbo].[InsertDataType] N'approximate_numbers', N'REAL', N'Real number type'
GO

EXECUTE [dbo].[InsertDataType] N'binary_data', N'BINARY', N'Simple binary data type'
EXECUTE [dbo].[InsertDataType] N'binary_data', N'VARBINARY', N'Variable binary data type'
EXECUTE [dbo].[InsertDataType] N'binary_data', N'IMAGE', N'Image type'
GO

EXECUTE [dbo].[InsertDataType] N'date_time', N'DATE', N'Simple date type'
EXECUTE [dbo].[InsertDataType] N'date_time', N'DATETIMEOFFSET', N'Date time offset type'
EXECUTE [dbo].[InsertDataType] N'date_time', N'DATETIME', N'Simple date time type'
EXECUTE [dbo].[InsertDataType] N'date_time', N'DATETIME2', N'Extended date time type'
EXECUTE [dbo].[InsertDataType] N'date_time', N'SMALLDATETIME', N'Small date time type'
EXECUTE [dbo].[InsertDataType] N'date_time', N'TIME', N'Simple time type'
GO

EXECUTE [dbo].[InsertDataType] N'other', N'CURSOR', N'Current set of recodrs type. Cannot be used for table field'
EXECUTE [dbo].[InsertDataType] N'other', N'TIMESTAMP', N'Timestamp type'
EXECUTE [dbo].[InsertDataType] N'other', N'HIERARCHYID', N'Hierarchy identifier type. Cannot be used for table field'
EXECUTE [dbo].[InsertDataType] N'other', N'UNIQUEIDENTIFIER', N'Unique identifier (GUID or UUID) type'
EXECUTE [dbo].[InsertDataType] N'other', N'SQL_VARIANT', N'Variant type (like ''undefined'' pointer)'
EXECUTE [dbo].[InsertDataType] N'other', N'XML', N'Extensible markup language object type'
EXECUTE [dbo].[InsertDataType] N'other', N'TABLE', N'Table, like cursor (or view) type. Cannot be used for table field'
GO

ALTER TABLE [dbo].[data_types] ADD [can_be_table_field] BIT DEFAULT 1
GO
UPDATE [dbo].[data_types] SET [can_be_table_field] = 1
UPDATE [dbo].[data_types] SET [can_be_table_field] = 0 WHERE
	[title] = N'CURSOR' OR
	[title] = N'HIERARCHYID' OR
	[title] = N'TABLE' OR
	[title] = N'DATETIME2'

GO

EXECUTE [dbo].[InsertDataType] N'references', N'1-1', N'Single-to-single reference'
EXECUTE [dbo].[InsertDataType] N'references', N'1-m', N'Single-to-multiple reference'
GO

INSERT INTO [dbo].[tables] ([title],[description]) VALUES ('tables_references','Definitions of references between tables')
GO
ALTER TABLE [dbo].[tables_references] ADD [id_parent_table] BIGINT NOT NULL
ALTER TABLE [dbo].[tables_references] ADD [id_child_table] BIGINT NOT NULL
ALTER TABLE [dbo].[tables_references] ADD CONSTRAINT [fk_parent_table]
	FOREIGN KEY ([id_parent_table])
	REFERENCES [dbo].[tables]([id])
ALTER TABLE [dbo].[tables_references] ADD CONSTRAINT [fk_child_table]
	FOREIGN KEY ([id_child_table])
	REFERENCES [dbo].[tables]([id])
GO

ALTER TABLE [dbo].[tables_fields] ADD [id_data_type_category]  BIGINT NOT NULL
ALTER TABLE [dbo].[tables_fields] ADD [id_data_type]  BIGINT NOT NULL
ALTER TABLE [dbo].[tables_fields] ADD [size]  TINYINT DEFAULT NULL
ALTER TABLE [dbo].[tables_fields] ADD [description]  NTEXT DEFAULT NULL
ALTER TABLE [dbo].[tables_fields] ADD [default_value]  SQL_VARIANT DEFAULT NULL
ALTER TABLE [dbo].[tables_fields] ADD [is_not_null] BIT NOT NULL DEFAULT 0
ALTER TABLE [dbo].[tables_fields] ADD [reference_by_table] NVARCHAR(255) DEFAULT NULL
ALTER TABLE [dbo].[tables_fields] ADD CONSTRAINT [fk_table_field_data_type_category_id] 
	FOREIGN KEY ([id_data_type_category])
	REFERENCES [dbo].[data_types_categories] ([id])
ALTER TABLE [dbo].[tables_fields] ADD CONSTRAINT [fk_table_field_data_type_id] 
	FOREIGN KEY ([id_data_type])
	REFERENCES [dbo].[data_types] ([id])
ALTER TABLE [dbo].[tables_fields] ADD [reference_to_id_table] BIGINT DEFAULT NULL
GO

EXECUTE [dbo].[DropObjectIfExists] 'f_GetDataTypeId', 'FN'
GO
CREATE FUNCTION [dbo].[f_GetDataTypeId] (
	@type_name NVARCHAR(255)
	)
RETURNS BIGINT
AS
BEGIN
	DECLARE @ret BIGINT
	SET @ret = -1
	SELECT @ret = [id] FROM [dbo].[data_types] WHERE [title] = @type_name
	RETURN @ret
END
GO

EXECUTE [dbo].[DropObjectIfExists] 'f_GetDataTypeName', 'FN'
GO
CREATE FUNCTION [dbo].[f_GetDataTypeName] (
	@type_id BIGINT
	)
RETURNS NVARCHAR(255)
AS
BEGIN
	DECLARE @ret NVARCHAR(255)
	SET @ret = N''
	SELECT @ret = [title] FROM [dbo].[data_types] WHERE id = @type_id
	RETURN @ret
END
GO

EXECUTE [dbo].[DropObjectIfExists] 'f_GetFieldId', 'FN'
GO
CREATE FUNCTION [dbo].[f_GetId] (
	@table_name NVARCHAR(255),
	@title NVARCHAR(255)
	)
RETURNS BIGINT
AS
BEGIN
	DECLARE @table_id BIGINT
	SELECT @table_id = [dbo].[f_GetTableId](@table_name)
	DECLARE @ret BIGINT
	SELECT @ret = [id] FROM [dbo].[tables_fields] WHERE [id_table] = @table_id AND [title] = @title
	RETURN @ret
END
GO

EXECUTE [dbo].[DropObjectIfExists] 'Add_Standart_Field', 'P'
GO
CREATE PROCEDURE [dbo].[Add_Standart_Field] 
	@table_name NVARCHAR(255),
	@field_name NVARCHAR(255),
	@field_type NVARCHAR(255),
	@is_not_null BIT,
	@field_size TINYINT,
	@def_val SQL_VARIANT
AS
BEGIN
	DECLARE @statement NVARCHAR(2000)
	SET @statement = 
	N'ALTER TABLE [dbo].[' + @table_name + N'] ADD [' + @field_name + '] ' + @field_type
	IF @is_not_null = 1
		SET @statement = @statement + N' NOT NULL '
	IF @field_size IS NOT NULL
		SET @statement = @statement + N'(' + CAST(@field_size AS NVARCHAR(16)) + N')'
	IF @def_val IS NOT NULL
		SET @statement = @statement + N' DEFAULT ' + CAST(@def_val AS NVARCHAR(1000))
	EXECUTE [dbo].[sp_executesql] @statement
	RETURN 0
END
GO

EXECUTE [dbo].[DropObjectIfExists] 'Add_SSReference_Field', 'P'
GO
CREATE PROCEDURE [dbo].[Add_SSReference_Field]
	@table_name NVARCHAR(255),
	@field_name NVARCHAR(255),
	@field_type NVARCHAR(255),	
	@field_ref_to NVARCHAR(255)	
AS
BEGIN
	DECLARE @statement NVARCHAR(2000)
	SET @statement = 
	N'ALTER TABLE [dbo].[' + @table_name + N'] ADD [' + @field_name + N'] BIGINT NOT NULL 
	ALTER TABLE [dbo].[' + @table_name + N'] ADD CONSTRAINT [fk_' + @field_name + N'_ssrefto_' + @field_ref_to + N'] 
		FOREIGN KEY [' + @field_name + N'] 
		REFERENCES [dbo].[' + @field_ref_to + N'] ([id])'
	EXECUTE [dbo].[sp_executesql] @statement
	RETURN 0
END
GO

EXECUTE [dbo].[DropObjectIfExists] 'Add_SMReference_Field', 'P'
GO
CREATE PROCEDURE [dbo].[Add_SMReference_Field]
	@table_name NVARCHAR(255),
	@field_name NVARCHAR(255),
	@field_type NVARCHAR(255),	
	@field_ref_to NVARCHAR(255)
AS
BEGIN	
	DECLARE @cr_t_name NVARCHAR(255)
	SET @cr_t_name =  N'[dbo].[relation_' + @table_name + N'_' + @field_ref_to + N']'
	DECLARE @cr_t NVARCHAR(2000)
	SET @cr_t = 
	N'CREATE TABLE ' + @cr_t_name + N'(
		[id] BIGINT IDENTITY PRIMARY KEY CLUSTERED, 
		[id_' + @field_ref_to + N'] BIGINT NOT NULL,
		[id_' + @table_name + N'] BIGINT NOT NULL)'
	PRINT @cr_t
	EXECUTE [dbo].[sp_executesql] @cr_t
	SET @cr_t = 
	N'ALTER TABLE ' + @cr_t_name + ' ADD CONSTRAINT [fk_ref_id_' + @field_ref_to + N'] 
		FOREIGN KEY ([id_' + @field_ref_to + N'])
		REFERENCES [dbo].[' + @field_ref_to + N']([id])'
	EXECUTE [dbo].[sp_executesql] @cr_t
	SET @cr_t = 
	N'ALTER TABLE ' + @cr_t_name + ' ADD CONSTRAINT [fk_ref_id_' + @table_name + N'] 
		FOREIGN KEY ([id_' + @table_name + N'])
		REFERENCES [dbo].[' + @table_name + N']([id])'
	EXECUTE [dbo].[sp_executesql] @cr_t
	DECLARE @id_field BIGINT
	SELECT @id_field = [dbo].[f_GetFieldId] (@table_name, @field_name)
	UPDATE [dbo].[tables_fields] SET [reference_by_table] = @cr_t_name WHERE [id] = @id_field
	ALTER TABLE [dbo].[tables_fields] DISABLE TRIGGER [trig_insert_field_add_field_to_table]
	DECLARE @id_ref_table BIGINT
	SELECT @id_ref_table = [dbo].[f_GetTableId] (@field_ref_to)
	DECLARE @id_cat BIGINT
	DECLARE @id_type BIGINT
	DECLARE @size TINYINT
	DECLARE @descr NVARCHAR(2000)
	DECLARE @def_val SQL_VARIANT
	DECLARE @is_not_null BIT
	DECLARE @ref_to_id_table BIGINT
	SELECT @ref_to_id_table = [dbo].[f_GetTableId] (@table_name)
	SELECT 
		@id_cat = [id_data_type_category],
		@id_type = [id_data_type],
		@size = [size],
		@descr = [description],
		@def_val = [default_value],
		@is_not_null = [is_not_null]
	FROM [dbo].[tables_fields] WHERE [id] = @id_field
	INSERT INTO [dbo].[tables_fields] (
		[title],
		[id_table],
		[id_data_type_category],
		[id_data_type],
		[size],
		[description],
		[default_value],
		[is_not_null],
		[reference_to_id_table],
		[reference_by_table]) VALUES (
		@field_name,
		@id_ref_table,
		@id_cat,
		@id_type,
		@size,
		@descr,
		@def_val,
		@is_not_null,
		@ref_to_id_table,
		@cr_t_name)
	ALTER TABLE [dbo].[tables_fields] ENABLE TRIGGER [trig_insert_field_add_field_to_table]	
END
GO

EXECUTE [dbo].[DropObjectIfExists] 'Add_Reference_Field', 'P'
GO
CREATE PROCEDURE [dbo].[Add_Reference_Field]
	@table_name NVARCHAR(255),
	@field_name NVARCHAR(255),
	@field_type NVARCHAR(255),	
	@field_ref_to NVARCHAR(255)
AS
BEGIN
	IF @field_type = N'1-1' EXECUTE [dbo].[Add_SSReference_Field] @table_name, @field_name, @field_ref_to
	ELSE IF @field_type = N'1-m' EXECUTE [dbo].[Add_SMReference_Field] @table_name, @field_name, @field_ref_to
	RETURN 0
END
GO

EXECUTE [dbo].[DropObjectIfExists] 'AddFieldToTable', 'P'
GO
CREATE PROCEDURE [dbo].[AddFieldToTable]
	@table_name NVARCHAR(255),
	@field_name NVARCHAR(255),	
	@field_cat_name NVARCHAR(255),
	@field_type NVARCHAR(255),
	@is_not_null BIT = 0,
	@field_size TINYINT = NULL,
	@def_val SQL_VARIANT = NULL,
	@field_ref_to NVARCHAR(255) = NULL
AS
BEGIN	
	IF @field_cat_name = N'exact_numbers' EXECUTE [dbo].[Add_Standart_Field] @table_name, @field_name, @field_type, @is_not_null, @field_size, @def_val
	ELSE IF @field_cat_name = N'unicode_strings' EXECUTE [dbo].[Add_Standart_Field] @table_name, @field_name, @field_type, @is_not_null, @field_size, @def_val
	ELSE IF @field_cat_name = N'strings' EXECUTE [dbo].[Add_Standart_Field] @table_name, @field_name, @field_type, @is_not_null, @field_size, @def_val
	ELSE IF @field_cat_name = N'approximate_numbers' EXECUTE [dbo].[Add_Standart_Field] @table_name, @field_name, @field_type, @is_not_null, @field_size, @def_val
	ELSE IF @field_cat_name = N'binary_data' EXECUTE [dbo].[Add_Standart_Field] @table_name, @field_name, @field_type, @is_not_null, @field_size, @def_val
	ELSE IF @field_cat_name = N'date_time' EXECUTE [dbo].[Add_Standart_Field] @table_name, @field_name, @field_type, @is_not_null, @field_size, @def_val
	ELSE IF @field_cat_name = N'other' EXECUTE [dbo].[Add_Standart_Field] @table_name, @field_name, @field_type, @is_not_null, @field_size, @def_val
	ELSE IF @field_cat_name = N'references' EXECUTE [dbo].[Add_Reference_Field] @table_name, @field_name, @field_type, @field_ref_to
	RETURN 0
END
GO
/*end*/

EXECUTE [dbo].[DropObjectIfExists] 'trig_insert_field_add_field_to_table', 'TR'
GO
CREATE TRIGGER [trig_insert_field_add_field_to_table] ON [dbo].[tables_fields]
FOR INSERT
AS
BEGIN
	DECLARE @table_name NVARCHAR(255)
	DECLARE @field_cat NVARCHAR(255)
	DECLARE @field_type NVARCHAR(255)
	DECLARE @field_name NVARCHAR(255)
	DECLARE @field_size TINYINT
	DECLARE @def_val SQL_VARIANT
	DECLARE @ref_to NVARCHAR(255)
	DECLARE @table_id BIGINT
	SELECT @table_id = [id_table] FROM INSERTED	
	SELECT @table_name = [dbo].[f_GetTableName] (@table_id)		
	SELECT @field_name = [title] FROM INSERTED	
	DECLARE @cat_id BIGINT
	SELECT @cat_id = [id_data_type_category] FROM INSERTED
	SELECT @field_cat = [dbo].[f_GetDataTypeCategoryName] (@cat_id)	
	DECLARE @type_id BIGINT
	SELECT @type_id = [id_data_type] FROM INSERTED
	SELECT @field_type = [dbo].[f_GetDataTypeName] (@type_id)	
	SELECT @field_size = [size] FROM INSERTED
	SELECT @def_val = [default_value] FROM INSERTED	
	DECLARE @ref_id BIGINT
	SELECT @ref_id = [reference_to_id_table] FROM INSERTED	
	SELECT @ref_to = [dbo].[f_GetTableName] (@ref_id)
	DECLARE @is_not_null BIT
	SELECT @is_not_null = [is_not_null] FROM INSERTED
	EXECUTE [dbo].[AddFieldToTable]
		@table_name,
		@field_name,
		@field_cat,
		@field_type,
		@is_not_null,
		@field_size,
		@def_val,
		@ref_to

END
GO
EXECUTE [dbo].[sp_settriggerorder] 'trig_insert_field_add_field_to_table', 'last', 'insert'
GO

ALTER TABLE [dbo].[tables] ADD [is_system] BIT NOT NULL DEFAULT 0
ALTER TABLE [dbo].[tables] ADD [is_editable] BIT NOT NULL DEFAULT 1
GO

UPDATE [dbo].[tables] SET [is_system] = 1
UPDATE [dbo].[tables] SET [is_editable] = 0
GO

/*end*/
/*

use master
drop database common_db

*/